-- Designed by Toshio Iwata at DIGITALFILTER.COM 2015/09/22

LIBRARY ieee;
USE ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;

ENTITY tmds_encoder IS
  PORT(  
    clk      : IN  STD_LOGIC;                     
    disp_ena : IN  STD_LOGIC;                     
    control  : IN  STD_LOGIC_VECTOR(1 DOWNTO 0);  
    d_in     : IN  STD_LOGIC_VECTOR(7 DOWNTO 0);  
    q_out    : OUT STD_LOGIC_VECTOR(9 DOWNTO 0)); 
END tmds_encoder;

ARCHITECTURE logic OF tmds_encoder IS
  
  component fulladd4dvi 
	  port ( A, B, CIN : in std_logic;
	        SUM, COUT : out std_logic );
  end component;

  SIGNAL q_m       : STD_LOGIC_VECTOR(8 DOWNTO 0); --internal data
  SIGNAL one_count_d_in : signed(7 downto 0); -- RANGE 0 TO 8;         --number of ones in d_in
  SIGNAL one_count_q_m  : signed(7 downto 0); -- INTEGER RANGE 0 TO 8;         --number of ones in q_m
  SIGNAL diff_q_m  : signed(7 downto 0); -- INTEGER RANGE -8 TO 8;        --number of ones minus the number of zeros in q_m
  SIGNAL disparity : signed(7 downto 0); -- INTEGER RANGE -16 TO 15;      --disparity
  SIGNAL sum_a0, sum_b0, sum_c0, sum_d0, sum_e0, sum_f0, sum_g0 : std_logic;
  SIGNAL cout_a0, cout_b0, cout_c0, cout_d0, cout_e0, cout_f0, cout_g0 : std_logic;
  SIGNAL sum_a1, sum_b1, sum_c1, sum_d1, sum_e1, sum_f1, sum_g1 : std_logic;
  SIGNAL cout_a1, cout_b1, cout_c1, cout_d1, cout_e1, cout_f1, cout_g1 : std_logic;
BEGIN

  --count the ones in d_in
  COMP_A0 : fulladd4dvi port map ( d_in(0), d_in(1), d_in(2), sum_a0, cout_a0 );
  COMP_B0 : fulladd4dvi port map ( d_in(3), d_in(4), d_in(5), sum_b0, cout_b0 );
  COMP_C0 : fulladd4dvi port map ( d_in(6), d_in(7), '0', sum_c0, cout_c0 );
  
  COMP_D0 : fulladd4dvi port map ( sum_a0, sum_b0, sum_c0, sum_d0, cout_d0 );
  COMP_E0 : fulladd4dvi port map ( cout_a0, cout_b0, cout_c0, sum_e0, cout_e0 );

  COMP_F0 : fulladd4dvi port map ( cout_d0, sum_e0, '0', sum_f0, cout_f0 );
  
  COMP_G0 : fulladd4dvi port map ( cout_f0, cout_e0, '0', sum_g0, cout_g0 );
  
  one_count_d_in(0) <= sum_d0;
  one_count_d_in(1) <= sum_f0;
  one_count_d_in(2) <= sum_g0;
  one_count_d_in(3) <= cout_g0;
  one_count_d_in(4) <= '0';
  one_count_d_in(5) <= '0';
  one_count_d_in(6) <= '0';
  one_count_d_in(7) <= '0';

  --process interval data to minimize transitions
  PROCESS(d_in, q_m, one_count_d_in)
  BEGIN
    IF(one_count_d_in > 4 OR (one_count_d_in = 4 AND d_in(0) = '0')) THEN
      q_m(0) <= d_in(0);
      q_m(1) <= q_m(0) XNOR d_in(1);
      q_m(2) <= q_m(1) XNOR d_in(2);
      q_m(3) <= q_m(2) XNOR d_in(3);
      q_m(4) <= q_m(3) XNOR d_in(4);
      q_m(5) <= q_m(4) XNOR d_in(5);
      q_m(6) <= q_m(5) XNOR d_in(6);
      q_m(7) <= q_m(6) XNOR d_in(7);
      q_m(8) <= '0';
    ELSE
      q_m(0) <= d_in(0);
      q_m(1) <= q_m(0) XOR d_in(1);
      q_m(2) <= q_m(1) XOR d_in(2);
      q_m(3) <= q_m(2) XOR d_in(3);
      q_m(4) <= q_m(3) XOR d_in(4);
      q_m(5) <= q_m(4) XOR d_in(5);
      q_m(6) <= q_m(5) XOR d_in(6);
      q_m(7) <= q_m(6) XOR d_in(7);
      q_m(8) <= '1';
    END IF;
  END PROCESS;
  
  --count the ones in q_m
  COMP_A1 : fulladd4dvi port map ( q_m(0), q_m(1), q_m(2), sum_a1, cout_a1 );
  COMP_B1 : fulladd4dvi port map ( q_m(3), q_m(4), q_m(5), sum_b1, cout_b1 );
  COMP_C1 : fulladd4dvi port map ( q_m(6), q_m(7), '0', sum_c1, cout_c1 );
  
  COMP_D1 : fulladd4dvi port map ( sum_a1, sum_b1, sum_c1, sum_d1, cout_d1 );
  COMP_E1 : fulladd4dvi port map ( cout_a1, cout_b1, cout_c1, sum_e1, cout_e1 );

  COMP_F1 : fulladd4dvi port map ( cout_d1, sum_e1, '0', sum_f1, cout_f1 );
  
  COMP_G1 : fulladd4dvi port map ( cout_f1, cout_e1, '0', sum_g1, cout_g1 );
  
  one_count_q_m(0) <= sum_d1;
  one_count_q_m(1) <= sum_f1;
  one_count_q_m(2) <= sum_g1;
  one_count_q_m(3) <= cout_g1;
  one_count_q_m(4) <= '0';
  one_count_q_m(5) <= '0';
  one_count_q_m(6) <= '0';
  one_count_q_m(7) <= '0';
  
  diff_q_m <= one_count_q_m + one_count_q_m - 8;
  
  --determine output and new disparity
  PROCESS(clk)
  BEGIN
    IF(clk'EVENT AND clk = '1') THEN
      IF(disp_ena = '1') THEN  
        IF(disparity = 0 OR one_count_q_m = 4) THEN
          IF(q_m(8) = '0') THEN
            q_out <= NOT q_m(8) & q_m(8) & NOT q_m(7 DOWNTO 0);
            disparity <= disparity - diff_q_m;
          ELSE
            q_out <= NOT q_m(8)& q_m(8 DOWNTO 0);
            disparity <= disparity + diff_q_m;
          END IF;
        ELSE
          IF((disparity > 0 AND one_count_q_m > 4) OR (disparity < 0 AND one_count_q_m < 4)) THEN
            q_out <= '1' & q_m(8) & NOT q_m(7 DOWNTO 0);
            IF(q_m(8) = '0') THEN
              disparity <= disparity - diff_q_m;
            ELSE
              disparity <= disparity - diff_q_m + 2;
            END IF;
          ELSE
            q_out <= '0' & q_m(8 DOWNTO 0);
            IF(q_m(8) = '0') THEN
              disparity <= disparity + diff_q_m - 2;
            ELSE
              disparity <= disparity + diff_q_m;
            END IF;
          END IF;
        END IF;
      ELSE
        CASE control IS
          WHEN "00" => q_out <= "1101010100";
          WHEN "01" => q_out <= "0010101011";
          WHEN "10" => q_out <= "0101010100";
          WHEN "11" => q_out <= "1010101011";
          WHEN OTHERS => NULL;
        END CASE;
        disparity <= (others => '0');
      END IF;
    END IF;
  END PROCESS;
  
END logic;
